--[[
						     ----o----(||)----o----(||)----o----


                                                           #########  #  #######
                                                               #      #  #      ##
                                                               #      #  #        #
                                                               #      #  #      ##
                                                               #      #  #######
                                                               #      #  #
                                                               #      #  #
                                                               #      #  #
                                                               #      #  #


						     ----o----(||)----o----(||)----o----

							   v1.10 25th October 2016
							  Copyright (C) Chris Birch

						     ----o----(||)----o----(||)----o----

]]
--------------------------------------------------------------------------------------------------------------------------------------------

local gmatch = string.gmatch
local lower = string.lower
local match = string.match
local print = print
local select = select
local strsplit = strsplit
local sub = string.sub
local tonumber = tonumber
local tostring = tostring

local EnumerateFrames = EnumerateFrames
local GetBuybackItemLink = GetBuybackItemLink
local GetCurrencyListLink = GetCurrencyListLink
local GetItemSpell = GetItemSpell
local GetMerchantItemLink = GetMerchantItemLink
local GetPetInfoByPetID = C_PetJournal.GetPetInfoByPetID
local GetPetInfoBySpeciesID = C_PetJournal.GetPetInfoBySpeciesID
local GetSpellInfo = GetSpellInfo
local GetTradeSkillReagentItemLink = GetTradeSkillReagentItemLink
local MouseIsOver = MouseIsOver
local UnitAura = UnitAura
local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff
local UnitGUID = UnitGUID

local prefix = "\124cFF0070DETip\124cFF00B7EB "

--------------------------------------------------------------------------------------------------------------------------------------------
-- Chat linked Reference - Companion and Battle Pets (I also show the IDs in chat)
--------------------------------------------------------------------------------------------------------------------------------------------

local savedBattlePetID, savedSpeciesID;

local function Tip_FloatingBattlePetShow( self, _, _, _, _, _, _, battlePetID )

	if ( savedBattlePetID ~= FloatingBattlePetTooltip.battlePetID ) and ( savedSpeciesID ~= FloatingBattlePetTooltip.speciesID ) then

		if ( tostring( FloatingBattlePetTooltip.battlePetID ) == "0000000000000000" ) then
			local _, _, _, npcID, _, _, _, _, _, _, _, displayID = GetPetInfoBySpeciesID( FloatingBattlePetTooltip.speciesID )
			print( "Species ID: \124cFF93CCEA".. FloatingBattlePetTooltip.speciesID.. "\n"..
				"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
				"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. tostring( npcID or "nil" ) )
		else
			local speciesID, _, _, _, _, displayID, _, _, _, _, npcID = GetPetInfoByPetID( FloatingBattlePetTooltip.battlePetID )
			print( "Species ID: \124cFF93CCEA".. speciesID.. "\n"..
				"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
				"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. tostring( npcID or "nil" ).. "\n"..
				"\124cFF00B7EBPet ID: \124cFF93CCEA".. FloatingBattlePetTooltip.battlePetID.. "\n" )
		end
		savedBattlePetID = FloatingBattlePetTooltip.battlePetID
		savedSpeciesID = FloatingBattlePetTooltip.speciesID
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Anything supported by GameTooltipSetHyperlink
--	* Artifact Relics
--------------------------------------------------------------------------------------------------------------------------------------------

local function Tip_GameTooltipSetHyperlink( self, link )

	local idType, spellItemID = match( link, "(%w+):(%d+)" )
	if ( TipDB.blankPrefixLine == true ) then
		GameTooltip:AddLine( "\n".. prefix.. sub( idType, 2, 2):upper().. sub( idType, 3 ).. " ID: \124cFF93CCEA".. spellItemID )
	else
		GameTooltip:AddLine( prefix.. sub( idType, 2, 2):upper().. sub( idType, 3 ).. " ID: \124cFF93CCEA".. spellItemID )
	end
	GameTooltip:Show()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Chat linked References for anything supported by ItemRefTooltipSetHyperlink
--------------------------------------------------------------------------------------------------------------------------------------------

local function Tip_ItemRefTooltipSetHyperlink( self, link )

	local idType, spellItemID = match( link, "(%w+):(%d+)" )
	if ( TipDB.blankPrefixLine == true ) then
		ItemRefTooltip:AddLine( "\n".. prefix.. sub( idType, 1, 1):upper().. sub( idType, 2 ).. " ID: \124cFF93CCEA".. spellItemID )
	else
		ItemRefTooltip:AddLine( prefix.. sub( idType, 1, 1):upper().. sub( idType, 2 ).. " ID: \124cFF93CCEA".. spellItemID )
	end
	ItemRefTooltip:Show()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Pet Journal Pet Card
--------------------------------------------------------------------------------------------------------------------------------------------

local function Tip_PetJournalUpdatePetCard( self )

	if ( PetJournalPetCard.petID == nil ) then
		local _, _, _, npcID, _, _, _, _, _, _, _, displayID = GetPetInfoBySpeciesID( PetJournalPetCard.speciesID )
		if ( TipDB.blankPrefixLine == true ) then
			GameTooltip:AddLine( "\n".. prefix.. "Species ID: \124cFF93CCEA".. PetJournalPetCard.speciesID.. "\n"..
					"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
					"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. npcID )

		else
			GameTooltip:AddLine( prefix.. "Species ID: \124cFF93CCEA".. PetJournalPetCard.speciesID.. "\n"..
					"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
					"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. npcID )
		end
	else
		local speciesID, _, _, _, _, displayID, _, _, _, _, npcID = GetPetInfoByPetID( PetJournalPetCard.petID )
		if ( TipDB.blankPrefixLine == true ) then
			GameTooltip:AddLine( "\n".. prefix.. "Species ID: \124cFF93CCEA".. speciesID.. "\n"..
					"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
					"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. npcID.. "\n"..
					"\124cFF00B7EBPet ID: \124cFF93CCEA".. PetJournalPetCard.petID.. "\n" )
		else
			GameTooltip:AddLine( prefix.. "Species ID: \124cFF93CCEA".. speciesID.. "\n"..
					"\124cFF00B7EBDisplay ID: \124cFF93CCEA".. displayID..
					"\124cFF00B7EB   NPC ID: \124cFF93CCEA".. npcID.. "\n"..
					"\124cFF00B7EBPet ID: \124cFF93CCEA".. PetJournalPetCard.petID.. "\n" )
		end
	end
	GameTooltip:Show()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Add Line / Spell Consolidation / Item Consolidation - added v1.02
--------------------------------------------------------------------------------------------------------------------------------------------

local function AddLine( type, id )
	local line, found = "", false
	local compareLine = prefix.. type.. " ID:"
	for i = GameTooltip:NumLines(), 3, -1 do
		line = _G["GameTooltipTextLeft"..i ]:GetText()
		if ( line ~= nil ) then
			if ( match( line, compareLine ) ) then
				found = true
				break
			end
		end
	end
	if ( found == false ) then
		compareLine = ( TipDB.blankPrefixLine == true ) and ( "\n".. compareLine ) or compareLine
		GameTooltip:AddLine( compareLine.. " \124cFF93CCEA".. id )
		GameTooltip:Show()
	end
end

local function GetItemFromLink( link )
	local itemLink = link or select( 2, GameTooltip:GetItem() )
	if ( itemLink ~= nil ) then
		local spellItemID = tonumber( match( itemLink, "item:(%d+):" ) )
		if ( spellItemID ~= nil ) then
			AddLine( "Item", spellItemID )
		end
	end
end

local function GetSpellFromTooltip()
	local spellItemID = select( 3, GameTooltip:GetSpell() )
	if ( spellItemID ~= nil ) then
		AddLine( "Spell", spellItemID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Action Bar Button
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetAction( self, buttonNum )

	local spellItemID = select( 3, GameTooltip:GetSpell() )
	if ( spellItemID == nil ) then
		GetItemFromLink()
	else
		AddLine( "Spell", spellItemID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Artifact Pane - Traits
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetArtifactPowerByID( self, powerID )
	local spellItemID = C_ArtifactUI.GetPowerInfo( powerID )
	AddLine( "Spell", spellItemID )
	AddLine( "Artifact Power", powerID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Auctions - Items for sale, whether by the player or someone else
--------------------------------------------------------------------------------------------------------------------------------------------

-- Note: Recipes strangely return 0 yet do not show up for spells either

function Tip_TooltipSetAuctionItem( self, type, index )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Auctions - Items being setup for sale and not yet posted for auction
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetAuctionSellItem( self )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Bank and Backpack Bags
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetBagItem( self, container, slot )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Merchant Buyback Item
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetBuybackItem( self, itemIndex )
	GetItemFromLink( GetBuybackItemLink( itemIndex ) )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Currency
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetCurrencyToken( self, index )
	local currencyID = match( GetCurrencyListLink( index ), "currency:(%d+)" )
	AddLine( "Currency", currencyID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Gem Socket (already set)
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetExistingSocketGem( self, index )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Guild Bank
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetGuildBankItem( self, parm1, parm2 )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Heirloom tab of the Collections frame
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetHeirloomByItemID( self, spellItemID )
	if ( spellItemID ~= nil ) then
		AddLine( "Spell", spellItemID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Inbox & Open Mail Attachment items
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetInboxItem( self, mailIndex, attachmentIndex )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Inventory Item of player/target Character Frame; Main Bank and Reagent Bank Frames
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetInventoryItem( self, unitID, slot )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Game Time Tokens
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetItemByID( self, itemID  )
	AddLine( "Item", itemID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Loot Frame Items
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetLootItem( self, slot )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Merchant Trade Item
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetMerchantItem( self, itemIndex )
	GetItemFromLink( GetMerchantItemLink( itemIndex ) )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Mounts pane of the Collections Frame
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetMountBySpellID( self, spellItemID )
	AddLine( "Spell", spellItemID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Quest Items Rewards (prior to accepting the quest) & Required Items (prior to handing in)
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestItem( self, type, itemIndex )
	GetItemFromLink( GetQuestItemLink( type, itemIndex ) )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Quest Log Items Rewards (an accepted quest with fixed and/or choice of rewards)
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestLogItem( self, type, itemIndex )

--print(tostring(self:GetOwner():GetName()))
--print(tostring(self:GetOwner().type))
--print(tostring(self:GetOwner().objectType))
--print(tostring(self:GetOwner().Name:GetText()))

	if ( self:GetOwner().objectType == "item" ) then
		GetItemFromLink( GetQuestLogItemLink( type, itemIndex ) )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Quest Log Reward Spell
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestLogRewardSpell( self  )
--
--	GetSpellFromTooltip()
--
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Special Use Quest Items (clickable and displayed on the right side of the UI)
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestLogSpecialItem( self )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Quest Reward Spell
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetQuestRewardSpell( self  )
--
--	GetSpellFromTooltip()
--
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Profession Skill List Pane - Results & Reagents
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetRecipeItem( self, recipeID, reagentIndex )

	local link = ( reagentIndex == nil ) and C_TradeSkillUI.GetRecipeItemLink( recipeID ) or
				C_TradeSkillUI.GetRecipeReagentItemLink( recipeID, reagentIndex )

	if ( link ~= nil ) then
		local idType, spellItemID = match( link, "|H(%w+):(%d+)" )
		if ( spellItemID ~= nil ) then
			if ( idType == "item" ) then
				AddLine( "Item", spellItemID )
			else
				AddLine( "Spell", spellItemID )	-- e.g. Leather Working's Primal Leg Reinforcements is "enchant"
			end
		end
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Send Mail Attachment items
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetSendMailItem( self, attachmentIndex )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Player's Stance Button
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetShapeshift( self, index )
	GetSpellFromTooltip()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Gem Socket (placed into socket but not yet "set")
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetSocketGem( self, index )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Spellbook & Professions Panes and Spellbook
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetSpellBookItem( self, spellbookIndex, bookType )
	GetSpellFromTooltip()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Specialisation Pane @ Talent Frame; Core Abilities pane @ Spellbook
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetSpellByID( self, spellItemID )
	GetSpellFromTooltip()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Talent Pane of the player/target Talent Frame
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetTalent( self, ... )
	GetSpellFromTooltip()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Toybox of the Collections Frame
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetToyByItemID( self, spellItemID )
	AddLine( "Item", spellItemID )
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Trainer Pane
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetTrainerService( self, index )
	GetSpellFromTooltip()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Targeted Units - when mouse over a unit
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_OnUpdate( self, elapsed )

	local owner = GameTooltip:GetOwner()
	if ( owner ~= nil ) and ( owner:GetName() == "UIParent" ) then
		local _, unitID = GameTooltip:GetUnit()
		if ( unitID ~= nil ) then
			local guID = UnitGUID( unitID )
			if ( guID ~= nil ) then
				local type, _, _, _, _, npcID, _ = strsplit( "-", guID )
				if ( type == "Player" ) or ( type == "Vignette" ) or ( npcID == nil) then
				else
					local line = _G["GameTooltipTextLeft"..GameTooltip:NumLines() ]:GetText()
					if ( line ~= prefix.. "NPC ID: \124cFF93CCEA".. npcID ) then
						AddLine( "NPC", npcID )
					end
				end
			end
		end
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Targeted Units - only when mouse over a portrait
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetUnit( self, unitID )

	local guID = UnitGUID( unitID )
	if ( guID ~= nil ) then
		local type, _, _, _, _, npcID, _ = strsplit( "-", guID )
		if ( type == "Player" ) or ( type == "Vignette" ) or ( npcID == nil ) then
		else
			AddLine( "NPC", npcID )
		end
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Player's Buffs and Debuffs
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetUnitAura( self, unitID, auraIndex, filter )

	local spellItemID = select( 11, UnitAura( unitID, auraIndex, filter ) )
	if ( spellItemID ~= nil ) then
		AddLine( "Spell", spellItemID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Target's Buffs
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetUnitBuff( self, unitID, auraIndex )

	local spellItemID = select( 11, UnitBuff( unitID, auraIndex ) )
	if ( spellItemID ~= nil ) then
		AddLine( "Spell", spellItemID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Target's Debuffs
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetUnitDebuff( self, unitID, auraIndex )

	local spellItemID = select( 11, UnitDebuff( unitID, auraIndex ) )
	if ( spellItemID ~= nil ) then
		AddLine( "Spell", spellItemID )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Void Storage item to be deposited
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetVoidDepositItem( self, index )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Void Storage item already deposited
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetVoidItem( self, one, locationIndex ) -- "one" because it always had a value of "1" in testing
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Void Storage item to be withdrawn
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_TooltipSetVoidWithdrawalItem( self, index )
	GetItemFromLink()
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Hooks for after most/all of the UI has been loaded
--------------------------------------------------------------------------------------------------------------------------------------------

function Tip_OnEvent( self, event, parm1 )

	if ( event == "ADDON_LOADED") and ( parm1 == "Blizzard_Collections" ) then
		PetJournalPetCardPetInfo:HookScript("OnEnter", Tip_PetJournalUpdatePetCard)
	elseif ( event == "ADDON_LOADED") and ( parm1 == "Tip" ) then
		hooksecurefunc( "FloatingBattlePet_Show", Tip_FloatingBattlePetShow )
		hooksecurefunc( ItemRefTooltip, "SetHyperlink", Tip_ItemRefTooltipSetHyperlink )
		hooksecurefunc( GameTooltip, "SetHyperlink", Tip_GameTooltipSetHyperlink )
	end
end

--------------------------------------------------------------------------------------------------------------------------------------------
-- Slash Command / Configuration Options
--------------------------------------------------------------------------------------------------------------------------------------------

SLASH_Tip1 = "/tip"

local function Slash( options )

	local firstParm, firstParm3, secondParm, secondParm3, thirdParm, thirdParm3, fourthParm, fourthParm3

	for v in gmatch( options, "%S+" ) do
		v = lower( v )
		if not firstParm3 then
			firstParm = v
			firstParm3 = sub( v,1,3 )
		elseif not secondParm3 then
			secondParm = v
			secondParm3 = sub( v,1,3 )
		elseif not thirdParm3 then
			thirdParm = v
			thirdParm3 = sub( v,1,3 )
		else
			fourthParm = v
			fourthParm3 = sub( v,1,3 )
			break
		end
	end

	if ( firstParm3 == "hel" ) or ( firstParm3 == "?" ) or ( firstParm3 == nil ) then
		print( "\124cFF0070DETip Chat Commands:\n"..
			"\124cFF00B7EB empty\124r  = Toggle Tip's empty first line off or on\n"..
			"\124cFF00B7EB ver\124r  = show version number\n")

	elseif ( firstParm3 == "emp" ) then
		TipDB.blankPrefixLine = not TipDB.blankPrefixLine
		print( "\124cFF0070DETip:\124r Empty prefix line is: \124cFF00B7EB".. tostring( ( TipDB.blankPrefixLine == true ) and "On" or "Off" ) )

	elseif ( firstParm3 == "ver" ) then
		local version = GetAddOnMetadata( "Tip", "Version" )
		print( "\124cFF0070DETip:\124r Version: \124cFF00B7EB".. version )
	end
end

SlashCmdList[ "Tip" ] = function( options ) Slash( options ) end