local EventFrame = CreateFrame("Frame");
EventFrame:RegisterEvent("VARIABLES_LOADED");

-- Make sure UIScale is a valid value
local function ValidateUIScale(val)
	if not val then val = 1; end
	val = tonumber(val);
	if val < 0.64 then
		val = 0.64;
	elseif val > 1 then
		val = 1;
	end
	return val;
end

-- Set UIScale
local function RefreshScale()
	local NewScale = 1;
	if nibUIScale_Char ~= nil then
		-- If Char variable not nil then use the Character UIScale
		NewScale = nibUIScale_Char;
	else
		-- Else use Account UIScale
		NewScale = nibUIScale_Account;
	end
	SetCVar("uiscale", NewScale);
end

-- Slash Command
SLASH_NIBUISCALE1 = "/nibuiscale";
local function handler(msg, editbox)
	local command, rest = msg:match("^(%S*)%s*(.-)$");
	if command == "account" then
	-- Account UIScale
		if rest ~= "" then
			-- If they entered a value then set new Account UIScale to that
			nibUIScale_Account = ValidateUIScale(rest);
			RefreshScale();
		end
		-- Display current Account UIScale
		print("-");
		print("UIScale for this Account is set to ["..tostring(nibUIScale_Account).."]");
	elseif command == "character" then
	-- Character UIScale
		if rest ~= "" then
			if rest == "off" then
				-- Disable Character UIScale and use Account UIScale
				nibUIScale_Char = nil;
				RefreshScale();
				print("-");
				print("This Character is now using the Account UIScale ["..tostring(nibUIScale_Account).."]");
			else
				-- Set new Character UIScale
				nibUIScale_Char = ValidateUIScale(rest);
				RefreshScale();
				print("-");
				print("UIScale for this Character is set to ["..tostring(nibUIScale_Char).."]");
			end			
		else
			if nibUIScale_Char ~= nil then
				-- Character using Character UIScale
				-- Display current Character UIScale
				print("-");
				print("UIScale for this Character is set to ["..tostring(nibUIScale_Char).."]");
				print("To set this Character back to using the Account UIScale, type: /nibuiscale character off");
			else
				-- Character using Account UIScale
				-- Display current Account UIScale
				print("-");
				print("This Character is using the Account UIScale ["..tostring(nibUIScale_Account).."]");
				print("To set this Character's UIScale individually, use: /nibuiscale character UIScale");
			end
		end
	else
		-- Display syntax help
		print("-");
		print("Syntax: /nibuiscale account|character UIscale|off");
		print("---");
		print("To set this Account's UIScale, use: /nibuiscale account UIScale");
		print("To set this Character's UIScale individually, use: /nibuiscale character UIScale");
		print("To set this Character to use the Account UIScale, use: /nibuiscale character off");
		print("---");
		print("UIScale can be any value between 0.64 and 1");
	end
end
SlashCmdList["NIBUISCALE"] = handler;

-- Events
local function EventHandler(self, event)
	if event == "VARIABLES_LOADED" then
		-- Make sure WoW is set to "Use UI Scale"
		SetCVar("useUiScale", 1);
		
		-- If Account UIScale hasn't been set, set it to the current UIScale
		if not nibUIScale_Account then nibUIScale_Account = GetCVar("uiscale"); end
		
		-- Validate UIScale and Refresh
		nibUIScale_Account = ValidateUIScale(nibUIScale_Account);
		RefreshScale();
	end	
end
EventFrame:SetScript("OnEvent", EventHandler);